Package: segtools
Type: Package
Title: What the package does (short line)
Version: 1.0
Date: 2012-06-27
Author: Who wrote it
Maintainer: Who to complain to <yourfault@somewhere.net>
Description: More about what it does (maybe more than one line)
License: What license is it under?
